@@ -4,17 +4,31 @@ use strict;
 use vars qw($VERSION %Types);
 use Carp;
 use DBI;
-$VERSION = '0.60';
+$VERSION = '0.70';
 
 # %Types is used for creating the data table if it doesn't exist already.
 # You may want to edit this.
-%Types = (                 # key          value          frozen    freeze
-	  'mysql'   => [qw/ varchar(127)  longblob       tinyint   1    /],
-	  'mSQL'    => [qw/ char(255)     char(255)      int       0    /],
-	  'Sybase'  => [qw/ varchar(255)  varbinary(255) tinyint   1    /],
-	  'default' => [qw/ varchar(255)  varbinary(255) tinyint   1    /],  #others
+%Types = (                 # key          value          frozen    freeze  keyless
+	  'mysql'   => [qw/ varchar(127)  longblob       tinyint   1          0 /],
+	  'mSQL'    => [qw/ char(255)     char(255)      int       0          0 /],
+	  'Sybase'  => [qw/ varchar(255)  varbinary(255) tinyint   1          0 /],
+	  'Oracle'  => [qw/ varchar(255)  varchar2(2000) integer   1          0 /],
+	  'CSV'     => [qw/ varchar(255)  varchar(255)   integer   1          1 /],
+	  'Informix'=> [qw/ nchar(120)    nchar(2000)    integer   0          0 /],
+	  'Solid'   => [qw/ varchar(255)  varbinary(2000) integer   1          0 /],  #others
+	  'default' => [qw/ varchar(255)  varchar(255)   integer   0          0 /],  #others
 	  );
 
+# list drivers that do run-time binding correctly
+my %CAN_BIND = (
+		'mysql' => 1,
+		'mSQL'  => 1,
+		'Oracle' => 1,
+		'Pg'    => 1,
+		'Informix' => 1,
+		'Solid'    => 1,
+	       );
+
 # Default options for the module
 my %DefaultOptions = (
 		      'table'      =>  'pdata',
@@ -47,15 +61,18 @@ sub TIEHASH {
 
     if (UNIVERSAL::isa($dsn,'DBI::db')) {
 	$dbh = $dsn;
-	$driver = ''; # XXX No way to find out?
+	$driver = $dsn->{Driver}{Name};
     } else {
 	$dsn = "dbi:$dsn" unless $dsn=~ /^dbi/;
 	($driver) = $dsn =~ /\w+:(\w+)/;
 
 
 	# Try to establish connection with data source.
+	delete $ENV{NLS_LANG} if $driver eq 'Oracle';  # allow 8 bit connections?
 	$dbh = DBI->connect($dsn,$opt->{user},$opt->{password},
 			    { AutoCommit=>$opt->{autocommit},
+			      PrintError=>0,
+			      ChopBlanks=>1,
 			      Warn=>0 }
 			    );
 	croak "TIEHASH: Can't open $dsn, $DBI::errstr" unless $dbh;
@@ -66,8 +83,11 @@ sub TIEHASH {
     # 1. The database driver supports binary types.
     # 2. The database table has a boolean field to indicate that a value is frozen.
     # 3. The Storable module is available.
-    my($canfreeze) = $Types{$driver}->[3];
-    my ($haveStorable) = eval 'require Storable;';
+    # we also check that "primary key" is recognized
+    my $db_features = $Types{$driver} || $Types{'default'};
+    my($canfreeze) = $db_features->[3];
+    my($keyless)   = $db_features->[4];
+    my($haveStorable) = eval 'require Storable;';
     Storable->import(qw/nfreeze thaw/) if $haveStorable;
     $canfreeze &&= $haveStorable;
 
@@ -80,10 +100,11 @@ sub TIEHASH {
     my $q = "select * from $opt->{table} where $opt->{key}=''";
     my $sth = $dbh->prepare($q);
     my $structure_ok = 0;
-    if ($sth->execute() ne '') { 
+    local($^W)=0; # uninitialized variable problem
+    if (defined($sth) && $sth->execute() ne '') { 
 	# At least the key field exists.  Check whether the others do too.
 	my(%field_names);
-	grep($field_names{$_}++,@{$sth->{NAME}});
+	grep($field_names{lc($_)}++,@{$sth->{NAME}});
 	$structure_ok++ if $field_names{$opt->{'value'}};
 	$canfreeze &&= $field_names{$opt->{'frozen'}};
     }
@@ -98,11 +119,11 @@ sub TIEHASH {
 	    
 	$dbh->do("drop table $opt->{table}") if $opt->{'drop'};
 
-	my($keytype,$valuetype,$frozentype) = @{$Types{$driver}};
-	($keytype,$valuetype,$frozentype)   = @{$Types{'default'}} unless $keytype;
-	my(@fields) = ("$opt->{key}    $keytype primary key",
+	my($keytype,$valuetype,$frozentype) = @{$db_features};
+	my(@fields) = ($keyless ? "$opt->{key}    $keytype" : "$opt->{key}    $keytype primary key",
 		       "$opt->{value}  $valuetype");
-	push(@fields,  "$opt->{frozen} $frozentype not null") if $canfreeze;
+	push(@fields,  ($keyless ? "$opt->{frozen} $frozentype" : "$opt->{frozen} $frozentype not null") )
+	  if $canfreeze;
 	$q = "create table $opt->{table} (" . join(',',@fields) . ")";
 	warn "$q\n" if $opt->{DEBUG};
 	$dbh->do($q) || do {
@@ -119,7 +140,8 @@ sub TIEHASH {
 	'value'     => $opt->{'value'},
 	'frozen'    => $opt->{'frozen'},
 	'canfreeze' => $canfreeze,
-	'broken'    => $driver eq 'mSQL',
+	'brokenselect'  => $driver eq 'mSQL',
+	'canbind'   => $CAN_BIND{$driver},
 	'DEBUG'     => $opt->{DEBUG},
     },$class;
 }
@@ -133,18 +155,13 @@ sub FETCH {
        if exists $self->{'cached_value'}->{$key};
 
     # create statement handler if it doesn't already exist.
-    unless ($self->{'fetch'}) {
-	my $cols = $self->{'canfreeze'} ? "$self->{'value'},$self->{'frozen'}" : $self->{'value'};
-	$self->{'fetch'} = $self->_prepare(<<END);
+    my $cols = $self->{'canfreeze'} ? "$self->{'value'},$self->{'frozen'}" : $self->{'value'};
+    my $sth = $self->_run_query('fetch',<<END,$key);
 select $cols from $self->{table} where $self->{key}=?
 END
     ;
-    }
-
-    $self->{'fetch'}->execute($key) || croak "Can't execute select: $DBI::errstr";
-    
-    my $result = $self->{'fetch'}->fetchrow_arrayref();
-    $self->{'fetch'}->finish;    
+    my $result = $sth->fetchrow_arrayref();
+    $sth->finish;    
     return undef unless $result;
     $self->{'canfreeze'} && $result->[1] ? thaw($result->[0]) : $result->[0];
 }
@@ -162,7 +179,7 @@ sub STORE {
     # for both database whose update statements return the number of rows updated,
     # and those (like mSQL) whose update statements don't.
     my($r);
-    if ($self->{'broken'}) {
+    if ($self->{'brokenselect'}) {
 	return $self->EXISTS($key) ? $self->_update($key,$value,$frozen)
 		                   : $self->_insert($key,$value,$frozen);
     }
@@ -172,44 +189,35 @@ sub STORE {
 
 sub DELETE {
     my($self,$key) = @_;
-    # Create the delete statement if it doesn't already exist
-    unless ($self->{'delete'}) {
-	$self->{'delete'} = $self->_prepare(<<END);
+    my $sth = $self->_run_query('delete',<<END,$key);
 delete from $self->{table} where $self->{key}=?
 END
     ;
-    }
-    $self->{'delete'}->execute($key);
-    croak "Database delete statement failed: $DBI::errstr" if $self->{'delete'}->err;        
+    croak "Database delete statement failed: $DBI::errstr" if $sth->err;        
+    $sth->finish;
     1;
 }
 
 sub CLEAR {
     my $self = shift;
     my $dbh = $self->{'dbh'};
-    # Check whether the clear statement handler already exists
-    unless ($self->{'clear'}) {
-	$self->{'clear'} = $self->_prepare("delete from $self->{table}");
-    }
-    $self->{'clear'}->execute();    
+    my $sth= $self->_prepare('clear',"delete from $self->{table}");
+    $sth->execute();    
     croak "Database delete all statement failed: $DBI::errstr" if $dbh->err;        
+    $sth->finish;
 }
 
 sub EXISTS {
     my($self,$key) = @_;
-
-    # Create the exists statement if it doesn't already exist
-    unless ($self->{'exists'}) {
-	$self->{'exists'} = $self->_prepare(<<END);
+    my $sth = $self->_run_query('exists',<<END,$key);
 select $self->{key} from $self->{table} where $self->{key}=?
 END
     ;
-    }
-
-    my $rows = $self->{'exists'}->execute($key);
-    croak "Database select statement failed: $DBI::errstr" if $self->{'exists'}->err;
-    $self->{'exists'}->finish;
-    $rows > 0;
+    croak "Database select statement failed: $DBI::errstr" unless $sth;
+    $sth->fetch;
+    my $rows = $sth->rows;
+    $sth->finish;
+    $rows >= 1;
 }
 
 sub FIRSTKEY {
@@ -217,30 +225,25 @@ sub FIRSTKEY {
     
     delete $self->{'cached_value'};
 
-    unless ($self->{'fetchkeys'}) {
-	$self->{'fetchkeys'} = $self->_prepare($self->{'canfreeze'} ? <<END1 : <<END2);
+    my $sth = $self->_prepare('fetchkeys',$self->{'canfreeze'} ? <<END1 : <<END2);
 select $self->{'key'},$self->{'value'},$self->{'frozen'} from $self->{'table'}
 END1
 select $self->{'key'},$self->{'value'} from $self->{'table'}
 END2
     ;
-    }
 
-    $self->{'fetchkeys'}->execute() 
-	|| croak "Can't execute select statement: $DBI::errstr";
-
-    # Save the statement handler for future reference
-    my $ref = $self->{'fetchkeys'}->fetch();
+    $sth->execute() || croak "Can't execute select statement: $DBI::errstr";
+    my $ref = $sth->fetch();
     return defined($ref) ? $ref->[0] : undef;
 }
 
 sub NEXTKEY {
     my $self = shift;
     # no statement handler defined, so nothing to iterate over
-    return wantarray ? () : undef unless $self->{'fetchkeys'};
-    my $r = $self->{'fetchkeys'}->fetch();
+    return wantarray ? () : undef unless my $sth = $self->{'fetchkeys'};
+    my $r = $sth->fetch();
     if (!$r) {
-	$self->{'fetchkeys'}->finish;
+	$sth->finish;
 	delete $self->{'cached_value'};
 	return wantarray ? () : undef;
     }
@@ -268,43 +271,62 @@ sub rollback {
 # utility routines
 sub _update {
     my ($self,$key,$value,$frozen) = @_;
-    my (@vars);
+    my ($sth);
     if ($self->{'canfreeze'}) {
-	$self->{'update'} ||= 
-	    $self->_prepare("update $self->{table} set $self->{value}=?,$self->{frozen}=? where $self->{key}=?");
-	@vars = ($value,$frozen,$key);
+	$sth = $self->_run_query('update',
+				 "update $self->{table} set $self->{value}=?,$self->{frozen}=? where $self->{key}=?",
+				 $value,$frozen,$key);
     } else {
-	$self->{'update'} ||= 
-	    $self->_prepare("update $self->{table} set $self->{value}=? where $self->{key}=?");
-	@vars = ($value,$key);
+	$sth = $self->_run_query('update',
+				 "update $self->{table} set $self->{value}=? where $self->{key}=?",
+				 $value,$key);
     }
-    my $r = $self->{'update'}->execute(@vars) 
-	|| croak "Update: $DBI::errstr";
-    $r > 0;
+    croak "Update: $DBI::errstr"  unless $sth;
+    $sth->rows > 0;
 }
 
 sub _insert {
     my ($self,$key,$value,$frozen) = @_;
-    my (@vars);
+    my ($sth);
     if ($self->{'canfreeze'}) {
-	$self->{'insert'} ||= 
-	    $self->_prepare("insert into $self->{table} ($self->{key},$self->{value},$self->{frozen}) values (?,?,?)");
-	@vars = ($key,$value,$frozen);
+	$sth = $self->_run_query('insert',
+				 "insert into $self->{table} ($self->{key},$self->{value},$self->{frozen}) values (?,?,?)",
+				 $key,$value,$frozen);
     } else {
-	$self->{'insert'} ||= 
-	    $self->_prepare("insert into $self->{table} ($self->{key},$self->{value}) values (?,?)");
-	@vars = ($key,$value);
+	$sth = $self->_run_query('insert',
+				 "insert into $self->{table} ($self->{key},$self->{value}) values (?,?)",
+				 $key,$value);
     }
-    $self->{'insert'}->execute(@vars) 
-	|| croak "Update: $DBI::errstr";
+    ($sth && $sth->rows) || croak "Update: $DBI::errstr";
 }
 
-sub _prepare ($$) {
-    my($self,$q) = @_;
-    warn $q,"\n" if $self->{DEBUG};
-    my $sth = $self->{'dbh'}->prepare($q);
-    croak qq/Problems preparing statement "$q": $DBI::errstr/ unless $sth;
-    $sth;
+sub _run_query {
+    my $self = shift;
+    my ($tag,$query,@bind_variables) = @_;
+    if ($self->{canbind}) {
+	my $sth = $self->_prepare($tag,$query);
+	return undef unless $sth->execute(@bind_variables);
+	return $sth;
+    }
+    # if we get here, then we can't bind, so we replace ? with escaped parameters
+    $query =~ s/\?/$self->{'dbh'}->quote(shift(@bind_variables))/eg;
+    my $sth = $self->{'dbh'}->prepare($query);
+    return undef unless $sth && $sth->execute;
+    return $sth;
+}
+
+sub _prepare ($$$) {
+    my($self,$tag,$q) = @_;
+    unless (exists($self->{$tag})) {
+	return undef unless $q;
+	warn $q,"\n" if $self->{DEBUG};
+	my $sth = $self->{'dbh'}->prepare($q);
+	croak qq/Problems preparing statement "$q": $DBI::errstr/ unless $sth;
+	$self->{$tag} = $sth;
+    } else {
+	$self->{$tag}->finish if $q;  # in case we forget
+    }
+    $self->{$tag};
 }
 
 1;
